# -*- coding: utf-8 -*-
"""
Wind Directions Command Line Interface
"""
import click
import logging
import os

from rex.utilities.loggers import init_mult
from rex.utilities.cli_dtypes import STR, INT
from rex.utilities.hpc import SLURM
from rex.utilities.utilities import get_class_properties

from reVX.config.wind_dirs import ProminentWindDirsConfig
from reVX.wind_dirs.prominent_wind_dirs import ProminentWindDirections
from reVX import __version__

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('--name', '-n', default='ProminentWindDirs', type=STR,
              show_default=True,
              help='Job name. Default is "ProminentWindDirs".')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, verbose):
    """
    Prominent Wind Directions Command Line Interface
    """
    ctx.ensure_object(dict)
    ctx.obj['NAME'] = name
    ctx.obj['VERBOSE'] = verbose


@main.command()
def valid_config_keys():
    """
    Echo the valid Prominent Wind Dirs config keys
    """
    click.echo(', '.join(get_class_properties(ProminentWindDirsConfig)))


def run_local(ctx, config):
    """
    Run ProminentWindDirections locally using config

    Parameters
    ----------
    ctx : click.ctx
        click ctx object
    config : reVX.config.prominent_wind_dirs.ProminentWindDirsConfig
        Prominent Wind Directions config object.
    """
    ctx.obj['NAME'] = config.name
    ctx.invoke(local,
               powerrose_h5_fpath=config.powerrose_h5_fpath,
               excl_fpath=config.excl_fpath,
               out_dir=config.dirout,
               agg_dset=config.agg_dset,
               tm_dset=config.tm_dset,
               resolution=config.resolution,
               excl_area=config.excl_area,
               max_workers=config.execution_control.max_workers,
               sites_per_worker=config.execution_control.sites_per_worker,
               log_dir=config.log_directory,
               verbose=config.log_level)


@main.command()
@click.option('--config', '-c', required=True,
              type=click.Path(exists=True),
              help='Filepath to ProminentWindDirections config json file.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config, verbose):
    """
    Run prominent wind directions from a config.
    """

    config = ProminentWindDirsConfig(config)

    if 'VERBOSE' in ctx.obj:
        if any((ctx.obj['VERBOSE'], verbose)):
            config._log_level = logging.DEBUG
    elif verbose:
        config._log_level = logging.DEBUG

    if config.execution_control.option == 'local':
        run_local(ctx, config)

    if config.execution_control.option == 'eagle':
        eagle(config)


@main.command()
@click.option('--powerrose_h5_fpath', '-prh5', required=True,
              type=click.Path(exists=True),
              help="Filepath to .h5 file containing powerrose data")
@click.option('--excl_fpath', '-excl', required=True,
              type=click.Path(exists=True),
              help="Filepath to exclusions h5 with techmap dataset.")
@click.option('--out_dir', '-o', required=True, type=click.Path(),
              help='Directory to dump output files')
@click.option('--agg_dset', '-ad', default='powerrose_100m', type=STR,
              show_default=True,
              help="Powerrose dataset to aggreate")
@click.option('--tm_dset', '-td', default='techmap_wtk', type=STR,
              show_default=True,
              help=("Dataset name in the techmap file containing the "
                    "exclusions-to-resource mapping data,"))
@click.option('--resolution', '-res', default=128, type=INT,
              show_default=True,
              help=("SC resolution, must be input in combination with gid. "
                    "Prefered option is to use the row / col slices to define "
                    "the SC point instead"))
@click.option('--excl_area', '-ea', default=0.0081, type=float,
              show_default=True,
              help="Area of an exclusion cell (square km)")
@click.option('--max_workers', '-mw', default=None, type=INT,
              show_default=True,
              help=("Number of cores to run summary on. None is all "
                    "available cpus"))
@click.option('--sites_per_worker', '-spw', default=1000, type=INT,
              show_default=True,
              help="Number of SC points to process on each parallel worker")
@click.option('--log_dir', '-log', default=None, type=STR,
              show_default=True,
              help='Directory to dump log files. Default is out_dir.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def local(ctx, powerrose_h5_fpath, excl_fpath, out_dir, agg_dset, tm_dset,
          resolution, excl_area, max_workers, sites_per_worker, log_dir,
          verbose):
    """
    Compute prominent wind directions on local hardware
    """
    sites_per_worker = sites_per_worker if sites_per_worker else 1000

    ctx.obj['OUT_DIR'] = out_dir
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    out_fpath = os.path.basename(powerrose_h5_fpath)
    out_fpath = out_fpath.replace('.h5',
                                  '_prominent_dir_{}.csv'
                                  .format(resolution))
    out_fpath = os.path.join(out_dir, out_fpath)

    if log_dir is None:
        log_dir = out_dir

    name = ctx.obj['NAME']
    if 'VERBOSE' in ctx.obj:
        verbose = any((ctx.obj['VERBOSE'], verbose))

    log_modules = [__name__, 'reVX', 'reV', 'rex']
    init_mult(name, log_dir, modules=log_modules, verbose=verbose)

    logger.info('Aggregating Prominent Wind Directions \n'
                'Outputs to be stored in: {}'.format(out_dir))

    ProminentWindDirections.run(powerrose_h5_fpath, excl_fpath,
                                agg_dset=agg_dset, tm_dset=tm_dset,
                                resolution=resolution, excl_area=excl_area,
                                max_workers=max_workers,
                                sites_per_worker=sites_per_worker,
                                out_fpath=out_fpath)


def get_node_cmd(config):
    """
    Get the node CLI call for prominent wind direction computation.

    Parameters
    ----------
    config : reVX.config.prominent_wind_dirs.ProminentWindDirsConfig
        Prominent Wind Directions config object.

    Returns
    -------
    cmd : str
        CLI call to submit to SLURM execution.
    """
    spw = config.execution_control.sites_per_worker
    args = ['-n {}'.format(SLURM.s(config.name)),
            'local',
            '-prh5 {}'.format(SLURM.s(config.powerrose_h5_fpath)),
            '-excl {}'.format(SLURM.s(config.excl_fpath)),
            '-o {}'.format(SLURM.s(config.dirout)),
            '-ad {}'.format(SLURM.s(config.agg_dset)),
            '-td {}'.format(SLURM.s(config.tm_dset)),
            '-res {}'.format(SLURM.s(config.resolution)),
            '-ea {}'.format(SLURM.s(config.excl_area)),
            '-mw {}'.format(SLURM.s(config.execution_control.max_workers)),
            '-spw {}'.format(SLURM.s(spw)),
            '-log {}'.format(SLURM.s(config.log_directory)),
            ]

    if config.log_level == logging.DEBUG:
        args.append('-v')

    cmd = ('python -m reVX.wind_dirs.prominent_wind_dirs_cli {}'
           .format(' '.join(args)))
    logger.debug('Submitting the following cli call:\n\t{}'.format(cmd))

    return cmd


def eagle(config):
    """
    Run prominent wind directions on Eagle HPC.

    Parameters
    ----------
    config : reVX.config.prominent_wind_dirs.ProminentWindDirsConfig
        Prominent Wind Directions config object.
    """

    cmd = get_node_cmd(config)
    name = config.name
    log_dir = config.log_directory
    stdout_path = os.path.join(log_dir, 'stdout/')

    slurm_manager = SLURM()

    logger.info('Running prominent wind directions computation on Eagle with '
                'node name "{}"'.format(name))
    out = slurm_manager.sbatch(cmd,
                               alloc=config.execution_control.allocation,
                               memory=config.execution_control.memory,
                               walltime=config.execution_control.walltime,
                               feature=config.execution_control.feature,
                               name=name, stdout_path=stdout_path,
                               conda_env=config.execution_control.conda_env,
                               module=config.execution_control.module)[0]
    if out:
        msg = ('Kicked off prominent wind direction calculation "{}" '
               '(SLURM jobid #{}) on Eagle.'
               .format(name, out))
    else:
        msg = ('Was unable to kick off prominent wind direction calculation '
               '"{}". Please see the stdout error messages'
               .format(name))

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running Prominent Wind Directions CLI')
        raise



# -*- coding: utf-8 -*-
"""
Transmission Least Cost Defaults
"""
import os

from reVX.least_cost_xmission.config.xmission_config import XmissionConfig

DEFAULTSDIR = os.path.dirname(os.path.realpath(__file__))

# Cost multipliers for medium and short lines
SHORT_MULT = 1.5
MEDIUM_MULT = 1.2

# Cut offs are originally in miles but are converted to kilometers
# Note that minimum tie-line distance may be greater than these distances
SHORT_CUTOFF = 3 * 5280 / 3.28084 / 1000
MEDIUM_CUTOFF = 10 * 5280 / 3.28084 / 1000

CELL_SIZE = 90  # meters, size of cell. Both dims must be equal

# Decimal % distance to buffer clipped cost raster by. This helps to find the
# cheapest path. Larger values will run slower. Should be greater than 1 or
# clip will be too small.
CLIP_RASTER_BUFFER = 1.05

# Number of load centers and sinks to connect to
NUM_LOAD_CENTERS = 1
NUM_SINKS = 1

# Costs multiplier for cells affected by transmission barriers This should be
# greater than the largest possible cost multiplier * number of cells to
# discourage from crossing
BARRIERS_MULT = 100

# Minimum transmission path length (km), anything less will be set to this
# value
MINIMUM_DIST_KM = 5.5

# Cost to connect to PCA load center. Completely synthetic and should only be
# used if no other connection is available
SINK_CONNECTION_COST = 1e11

# Cost to connect to a t-line that is outside of the cost calculation area.
# These are only to make reV happy and should never be connected to. Must
# be higher than the SINK cost
FAR_T_LINE_COST = SINK_CONNECTION_COST * 1000
FAR_T_LINE_LENGTH = 999991

# Connection costs for t-lines w/ inadequate voltage for tie line
LOW_VOLT_T_LINE_COST = FAR_T_LINE_COST * 2
LOW_VOLT_T_LINE_LENGTH = FAR_T_LINE_LENGTH + 1

HILL_MULT = 1
MTN_MULT = 1
HILL_SLOPE = 2
MTN_SLOPE = 8

METERS_IN_MILE = 1609.344

WATER_NLCD_CODE = 11
WATER_MULT = 10

# Default test multipliers for RI
TEST_DEFAULT_MULTS = {
    "land_use": {
        "cropland": 1,
        "forest": 4,
        "suburban": 5,
        "urban": 6,
        "wetland": 7
    },
    "slope": {
        "hill_mult": 2,
        "hill_slope": 2,
        "mtn_mult": 5,
        "mtn_slope": 8
    }
}

# Category strings for transmission features
TRANS_LINE_CAT = 'TransLine'
LOAD_CENTER_CAT = 'LoadCen'
SINK_CAT = 'PCALoadCen'
SUBSTATION_CAT = 'Substation'



"""
Load default configuration for tie-line cost determination

Mike Bannister
7/27/21
"""
import logging
import os
import numpy as np

from rex.utilities.utilities import safe_json_load

CONFIG_DIR = os.path.dirname(os.path.realpath(__file__))
CONFIG = {
    "base_line_costs": os.path.join(CONFIG_DIR, 'base_line_costs.json'),
    "iso_lookup": os.path.join(CONFIG_DIR, 'iso_lookup.json'),
    "iso_multipliers": os.path.join(CONFIG_DIR, 'iso_multipliers.json'),
    "land_use_classes": os.path.join(CONFIG_DIR, 'land_use_classes.json'),
    "new_substation_costs": os.path.join(CONFIG_DIR,
                                         'new_substation_costs.json'),

    "power_classes": os.path.join(CONFIG_DIR, 'power_classes.json'),
    "power_to_voltage": os.path.join(CONFIG_DIR, 'power_to_voltage.json'),
    "transformer_costs": os.path.join(CONFIG_DIR,
                                      'transformer_costs.json'),
    "upgrade_substation_costs": os.path.join(CONFIG_DIR,
                                             'upgrade_substation_costs.json')
}
logger = logging.getLogger(__name__)


class XmissionConfig(dict):
    """ Load default tie lines configuration """

    def __init__(self, config=None):
        """
        Use default values for all parameters if None.

        Parameters
        ----------
        config : str | dict, optional
            Path to json file containing Xmission cost configuration values,
            or jsonified dict of cost configuration values,
            or dictionary of configuration values,
            or dictionary of paths to config jsons,
            if None use defaults in './config.json'
        """
        super().__init__()

        self._load_config(CONFIG)
        self._load_config(config)

    def __getitem__(self, k):
        if k == 'reverse_iso':
            out = {v: k for k, v in self['iso_lookup'].items()}
        elif k == 'voltage_to_power':
            out = {v: k for k, v in self['power_to_voltage'].items()}
        elif k == 'line_power_to_classes':
            out = {v: k for k, v in self['power_classes'].items()}
        else:
            out = super().__getitem__(k)

        return out

    @staticmethod
    def _parse_cap_class(capacity):
        """
        Parse capacity class from input capacity which can be a number or a
        string

        Parameters
        ----------
        capacity : int | float | str
            Capacity of interest

        Returns
        -------
        cap_class : str
            Capacity class in format "{capacity}MW"
        """
        if not isinstance(capacity, str):
            cap_class = '{}MW'.format(int(capacity))
        elif not capacity.endswith('MW'):
            cap_class = capacity + 'MW'
        else:
            cap_class = capacity

        return cap_class

    def _load_config(self, config):
        """
        Load config

        Parameters
        ----------
        config : str | dict
            Path to json file containing Xmission cost configuration values,
            or jsonified dict of cost configuration values,
            or dictionary of configuration values,
            or dictionary of paths to config jsons
        """
        if config is not None:
            if isinstance(config, str):
                config = safe_json_load(config)

            if not isinstance(config, dict):
                msg = ('Xmission costs config must be a path to a json file, '
                       'a jsonified dictionary, or a dictionary, not: {}'
                       .format(config))
                logger.error(msg)
                raise ValueError(msg)

            for k, v in config.items():
                if v.endswith('.json'):
                    v = safe_json_load(v)

                self[k] = v

    def capacity_to_kv(self, capacity):
        """
        Convert capacity class to line voltage

        Parameters
        ----------
        capacity : int
            Capacity class in MW

        Returns
        -------
        kv : int
            Tie-line voltage in kv
        """
        cap_class = self._parse_cap_class(capacity)
        line_capacity = self['power_classes'][cap_class]
        kv = self['power_to_voltage'][str(line_capacity)]

        return int(kv)

    def kv_to_capacity(self, kv):
        """
        Convert line voltage to capacity class

        Parameters
        ----------
        kv : in
            Tie-line voltage in kv

        Returns
        -------
        capacity : int
            Capacity class in MW
        """
        line_capacity = self['voltage_to_power'][kv]
        capacity = self['line_power_to_classes'][line_capacity].strip("MW")

        return int(capacity)

    def sub_upgrade_cost(self, region, tie_line_voltage):
        """
        Extract substation upgrade costs in $ for given region and tie-line
        voltage rating

        Parameters
        ----------
        region : int
            Region code, used to extract ISO
        tie_line_voltage : int | str
            Tie-line voltage class in kV

        Returns
        -------
        int
            Substation upgrade cost
        """
        if not isinstance(tie_line_voltage, str):
            tie_line_voltage = str(tie_line_voltage)

        region = self['reverse_iso'][region]

        return self['upgrade_substation_costs'][region][tie_line_voltage]

    def new_sub_cost(self, region, tie_line_voltage):
        """
        Extract new substation costs in $ for given region and tie-line
        voltage rating

        Parameters
        ----------
        region : int
            Region code, used to extract ISO
        tie_line_voltage : int | str
            Tie-line voltage class in kV

        Returns
        -------
        int
            New substation cost
        """
        if not isinstance(tie_line_voltage, str):
            tie_line_voltage = str(tie_line_voltage)

        region = self['reverse_iso'][region]

        return self['new_substation_costs'][region][tie_line_voltage]

    def xformer_cost(self, feature_voltage, tie_line_voltage):
        """
        Extract transformer costs in $ for given region and tie-line
        voltage rating

        Parameters
        ----------
        feature_voltage : int
            Voltage of feature that tie-line is connecting to
        tie_line_voltage : int | str
            Tie-line voltage class in kV

        Returns
        -------
        int
            Transformer cost as $/MW
        """
        if not isinstance(tie_line_voltage, str):
            tie_line_voltage = str(tie_line_voltage)

        costs = self['transformer_costs'][tie_line_voltage]

        classes = np.array(sorted(map(int, costs)))
        valid_idx = np.where(classes >= feature_voltage)[0]
        if valid_idx.size:
            v_class = classes[valid_idx[0]]
        else:
            v_class = classes[-1]

        return costs[str(v_class)]



# -*- coding: utf-8 -*-
"""
Compute prominent wind directions from PowerRose
"""
from .mean_wind_dirs import MeanWindDirections
from .prominent_wind_dirs import ProminentWindDirections, row_col_indices



# -*- coding: utf-8 -*-
"""
Handler to convert exclusion to/from .h5 and .geotiff
"""
import h5py
import json
import logging
import numpy as np
import os
from pyproj.crs import CRS
import rasterio
from warnings import warn

from reV.handlers.exclusions import ExclusionLayers
from reV.handlers.outputs import Outputs

from reVX.handlers.geotiff import Geotiff
from reVX.utilities.exceptions import ExclusionsCheckError
from reVX.utilities.utilities import log_versions

logger = logging.getLogger(__name__)


class ExclusionsConverter:
    """
    Convert exclusion layers between .h5 and .tif (geotiff)
    """
    def __init__(self, excl_h5, hsds=False, chunks=(128, 128), replace=True):
        """
        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing or to contain exclusion layers
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        chunks : tuple, optional
            Chunk size of exclusions in .h5 and Geotiffs, by default (128, 128)
        replace : bool, optional
            Flag to replace existing layers if needed, by default True
        """
        log_versions(logger)
        self._excl_h5 = excl_h5
        self._hsds = hsds
        self._chunks = chunks
        self._replace = replace

    def __repr__(self):
        msg = "{} for {}".format(self.__class__.__name__, self._excl_h5)
        return msg

    def __getitem__(self, layer):
        """
        Parameters
        ----------
        layer : str
            Layer to extract data for

        Returns
        -------
        profile : dict
            Geotiff profile (attributes)
        values : ndarray
            Geotiff data
        """

        if layer not in self.layers:
            msg = "{} is not present in {}".format(layer, self._excl_h5)
            logger.error(msg)
            raise KeyError(msg)

        profile, values = self._extract_layer(self._excl_h5, layer,
                                              hsds=self._hsds)
        return profile, values

    def __setitem__(self, layer, geotiff):
        """
        Parameters
        ----------
        layer : str
            Layer to set
        geotiff : str
            Path to GeoTiff to load data from
        """
        self.geotiff_to_layer(layer, geotiff)

    @property
    def layers(self):
        """
        Available exclusion layers in .h5 file

        Returns
        -------
        layers : list
            Available layers in .h5 file
        """
        with ExclusionLayers(self._excl_h5, hsds=self._hsds) as exc:
            layers = exc.layers

        return layers

    @staticmethod
    def _init_h5(excl_h5, geotiff, chunks=(128, 128)):
        """
        Initialize exclusions .h5 file from geotiff:
        - Transfer profile, shape, and meta

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers
        geotiff : str
            Path to geotiff file
        chunks : tuple
            Chunk size of exclusions in Geotiff
        """
        logger.debug('\t- Initializing {} from {}'
                     .format(excl_h5, geotiff))
        with Geotiff(geotiff, chunks=chunks) as src:
            profile = src.profile
            shape = src.shape
            lat, lon = src.lat_lon
            logger.debug('\t- "profile", "meta", and "shape" extracted from {}'
                         .format(geotiff))

        try:
            with h5py.File(excl_h5, mode='w') as dst:
                dst.attrs['profile'] = json.dumps(profile)
                logger.debug('\t- Default profile:\n{}'.format(profile))
                dst.attrs['shape'] = shape
                logger.debug('\t- Default shape:\n{}'.format(shape))
                dst.attrs['chunks'] = chunks
                logger.debug('\t- Default chunks:\n{}'.format(chunks))

                dst.create_dataset('latitude', shape=lat.shape,
                                   dtype=np.float32, data=lat,
                                   chunks=chunks)
                logger.debug('\t- latitude coordiantes created')

                dst.create_dataset('longitude', shape=lon.shape,
                                   dtype=np.float32, data=lon,
                                   chunks=chunks)
                logger.debug('\t- longitude coordiantes created')
        except Exception:
            logger.exception("Error initilizing {}".format(excl_h5))
            if os.path.exists(excl_h5):
                os.remove(excl_h5)

    @staticmethod
    def _check_crs(baseline_crs, test_crs, ignore_keys=('no_defs',)):
        """
        Compare baseline and test crs values

        Parameters
        ----------
        baseline_crs : dict
            Baseline CRS to use a truth, must be a dict
        test_crs : dict
            Test CRS to compare with baseline, must be a dictionary
        ignore_keys : tuple
            Keys to not check

        Returns
        -------
        bad_crs : bool
            Flag if crs' do not match
        """
        bad_crs = False
        for k, true_v in baseline_crs.items():
            if k not in ignore_keys:
                test_v = test_crs.get(k, true_v)
                if true_v != test_v:
                    bad_crs = True

        return bad_crs

    @classmethod
    def _check_geotiff(cls, excl_h5, geotiff, chunks=(128, 128),
                       transform_atol=0.01, coord_atol=0.001):
        """
        Compare geotiff with exclusion layer, raise any errors

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers
        geotiff : str
            Path to geotiff file
        chunks : tuple
            Chunk size of exclusions in Geotiff
        transform_atol : float
            Absolute tolerance parameter when comparing geotiff transform data.
        coord_atol : float
            Absolute tolerance parameter when comparing new un-projected
            geotiff coordinates against previous coordinates.
        """
        with Geotiff(geotiff, chunks=chunks) as tif:
            with ExclusionLayers(excl_h5) as h5:
                if tif.bands > 1:
                    error = ('{} contains more than one band!'
                             .format(geotiff))
                    logger.error(error)
                    raise ExclusionsCheckError(error)

                if not np.array_equal(h5.shape, tif.shape):
                    error = ('Shape of exclusion data in {} and {} do not '
                             'match!'.format(geotiff, excl_h5))
                    logger.error(error)
                    raise ExclusionsCheckError(error)

                profile = h5.profile
                h5_crs = CRS.from_string(profile['crs']).to_dict()
                tif_crs = CRS.from_string(tif.profile['crs']).to_dict()
                bad_crs = cls._check_crs(h5_crs, tif_crs)
                if bad_crs:
                    error = ('Geospatial "crs" in {} and {} do not match!'
                             '\n {} !=\n {}'
                             .format(geotiff, excl_h5, tif_crs, h5_crs))
                    logger.error(error)
                    raise ExclusionsCheckError(error)

                if not np.allclose(profile['transform'],
                                   tif.profile['transform'],
                                   atol=transform_atol):
                    error = ('Geospatial "transform" in {} and {} do not '
                             'match!\n {} !=\n {}'
                             .format(geotiff, excl_h5, profile['transform'],
                                     tif.profile['transform']))
                    logger.error(error)
                    raise ExclusionsCheckError(error)

                lat, lon = tif.lat_lon
                if not np.allclose(h5.latitude, lat, atol=coord_atol):
                    error = ('Latitude coordinates {} and {} do not match to '
                             'within {} degrees!'
                             .format(geotiff, excl_h5, coord_atol))
                    logger.error(error)
                    raise ExclusionsCheckError(error)

                if not np.allclose(h5.longitude, lon, atol=coord_atol):
                    error = ('Longitude coordinates {} and {} do not match to '
                             'within {} degrees!'
                             .format(geotiff, excl_h5, coord_atol))
                    logger.error(error)
                    raise ExclusionsCheckError(error)

    @classmethod
    def _parse_tiff(cls, geotiff, excl_h5=None, chunks=(128, 128),
                    check_tiff=True, transform_atol=0.01, coord_atol=0.001):
        """
        Extract exclusion layer from given geotiff, compare with excl_h5
        if provided

        Parameters
        ----------
        geotiff : str
            Path to geotiff file
        excl_h5 : str, optional
            Path to .h5 file containing exclusion layers, by default None
        chunks : tuple, optional
            Chunk size of exclusions in Geotiff, by default (128, 128)
        check_tiff : bool, optional
            Flag to check tiff profile and coordinates against exclusion .h5
            profile and coordinates, by default True
        transform_atol : float, optional
            Absolute tolerance parameter when comparing geotiff transform data,
            by default 0.01
        coord_atol : float, optional
            Absolute tolerance parameter when comparing new un-projected
            geotiff coordinates against previous coordinates, by default 0.001

        Returns
        -------
        profile : dict
            Geotiff profile (attributes)
        values : ndarray
            Geotiff data
        """
        if excl_h5 is not None and check_tiff:
            cls._check_geotiff(excl_h5, geotiff, chunks=chunks,
                               transform_atol=transform_atol,
                               coord_atol=coord_atol)

        with Geotiff(geotiff, chunks=chunks) as tif:
            profile, values = tif.profile, tif.values

        return profile, values

    @staticmethod
    def _write_layer(excl_h5, layer, profile, values, chunks=(128, 128),
                     description=None, scale_factor=None):
        """
        Write exclusion layer to .h5 file

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers
        layer : str
            Dataset name in .h5 file
        profile : dict
            Geotiff profile (attributes)
        values : ndarray
            Geotiff data
        chunks : tuple
            Chunk size of dataset in .h5 file
        description : str
            Description of exclusion layer
        scale_factor : int | float, optional
            Scale factor to use to scale geotiff data when added to the .h5
            file, by default None
        """
        if len(chunks) < 3:
            chunks = (1, ) + chunks

        if values.ndim < 3:
            values = np.expand_dims(values, 0)

        with h5py.File(excl_h5, mode='a') as f:
            if layer in f:
                ds = f[layer]
                ds[...] = values
                logger.debug('\t- {} values replaced'.format(layer))
            else:
                ds = f.create_dataset(layer, shape=values.shape,
                                      dtype=values.dtype, chunks=chunks,
                                      data=values)
                logger.debug('\t- {} created and loaded'.format(layer))

            ds.attrs['profile'] = json.dumps(profile)
            logger.debug('\t- Unique profile for {} added:\n{}'
                         .format(layer, profile))
            if description is not None:
                ds.attrs['description'] = description
                logger.debug('\t- Description for {} added:\n{}'
                             .format(layer, description))

            if scale_factor is not None:
                ds.attrs['scale_factor'] = scale_factor
                logger.debug('\t- scale_factor for {} added:\n{}'
                             .format(layer, scale_factor))

    @classmethod
    def _geotiff_to_h5(cls, excl_h5, layer, geotiff, chunks=(128, 128),
                       check_tiff=True, transform_atol=0.01, coord_atol=0.001,
                       description=None, scale_factor=None, dtype='int16'):
        """
        Transfer geotiff exclusions to h5 confirming they match existing layers

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers
        layer : str
            Layer to extract
        geotiff : str
            Path to geotiff file
        chunks : tuple, optional
            Chunk size of exclusions in Geotiff, by default (128, 128)
        check_tiff : bool, optional
            Flag to check tiff profile and coordinates against exclusion .h5
            profile and coordinates, by default True
        transform_atol : float, optional
            Absolute tolerance parameter when comparing geotiff transform data,
            by default 0.01
        coord_atol : float, optional
            Absolute tolerance parameter when comparing new un-projected
            geotiff coordinates against previous coordinates, by default 0.001
        description : str, optional
            Description of exclusion layer, by default None
        scale_factor : int | float, optional
            Scale factor to use to scale geotiff data when added to the .h5
            file, by default None
        dtype : str, optional
            Dtype to save geotiff data as in the .h5 file. Only used when
            'scale_factor' is not None, by default 'int16'
        """
        logger.debug('\t- {} being extracted from {} and added to {}'
                     .format(layer, geotiff, os.path.basename(excl_h5)))

        profile, values = cls._parse_tiff(
            geotiff, excl_h5=excl_h5, chunks=chunks, check_tiff=check_tiff,
            transform_atol=transform_atol, coord_atol=coord_atol)

        if scale_factor is not None:
            attrs = {'scale_factor': scale_factor}
            values = Outputs._check_data_dtype(layer, values, dtype,
                                               attrs=attrs)

        cls._write_layer(excl_h5, layer, profile, values,
                         chunks=chunks, description=description,
                         scale_factor=scale_factor)

    @staticmethod
    def _write_geotiff(geotiff, profile, values):
        """
        Write values to geotiff with given profile

        Parameters
        ----------
        geotiff : str
            Path to geotiff file to save data to
        profile : dict
            Geotiff profile (attributes)
        values : ndarray
            Geotiff data
        """
        out_dir = os.path.dirname(geotiff)
        if not os.path.exists(out_dir):
            logger.debug("Creating {}".format(out_dir))
            os.makedirs(out_dir)

        if values.shape[0] != 1:
            values = np.expand_dims(values, 0)

        dtype = values.dtype.name
        profile['dtype'] = dtype
        if np.issubdtype(dtype, np.integer):
            dtype_max = np.iinfo(dtype).max
        else:
            dtype_max = np.finfo(dtype).max

        profile['nodata'] = dtype_max

        with rasterio.open(geotiff, 'w', **profile) as f:
            f.write(values)
            logger.debug('\t- {} created'.format(geotiff))

    @classmethod
    def _extract_layer(cls, excl_h5, layer, geotiff=None, hsds=False):
        """
        Extract given layer from exclusions .h5 file and write to geotiff .tif

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers
        layer : str
            Layer to extract
        geotiff : str
            Path to geotiff file
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS

        Returns
        -------
        profile : dict
            Geotiff profile (attributes)
        values : ndarray
            Geotiff data
        """
        logger.debug('\t - Extracting {} from {}'
                     .format(layer, os.path.basename(excl_h5)))
        with ExclusionLayers(excl_h5, hsds=hsds) as f:
            profile = f.get_layer_profile(layer)
            values = f.get_layer_values(layer)

        if geotiff is not None:
            logger.debug('\t- Writing {} to {}'.format(layer, geotiff))
            cls._write_geotiff(geotiff, profile, values)

        return profile, values

    def geotiff_to_layer(self, layer, geotiff, check_tiff=True,
                         transform_atol=0.01, coord_atol=0.001,
                         description=None, scale_factor=None, dtype='int16'):
        """
        Transfer geotiff exclusions to h5 confirming they match existing layers

        Parameters
        ----------
        layer : str
            Layer to extract
        geotiff : str
            Path to geotiff file
        check_tiff : bool, optional
            Flag to check tiff profile and coordinates against exclusion .h5
            profile and coordinates, by default True
        transform_atol : float, optional
            Absolute tolerance parameter when comparing geotiff transform data,
            by default 0.01
        coord_atol : float, optional
            Absolute tolerance parameter when comparing new un-projected
            geotiff coordinates against previous coordinates, by default 0.001
        description : str, optional
            Description of exclusion layer, by default None
        scale_factor : int | float, optional
            Scale factor to use to scale geotiff data when added to the .h5
            file, by default None
        dtype : str, optional
            Dtype to save geotiff data as in the .h5 file. Only used when
            'scale_factor' is not None, by default 'int16'
        """
        if not os.path.exists(self._excl_h5):
            self._init_h5(self._excl_h5, geotiff, chunks=self._chunks)

        if layer in self.layers:
            msg = ("{} is already present in {}"
                   .format(layer, self._excl_h5))
            if self._replace:
                msg += " and will be replaced"
                logger.warning(msg)
                warn(msg)
            else:
                msg += ", to 'replace' set to True"
                logger.error(msg)
                raise KeyError(msg)

        self._geotiff_to_h5(self._excl_h5, layer, geotiff,
                            chunks=self._chunks,
                            check_tiff=check_tiff,
                            transform_atol=transform_atol,
                            coord_atol=coord_atol,
                            description=description,
                            scale_factor=scale_factor,
                            dtype=dtype)

    def layer_to_geotiff(self, layer, geotiff):
        """
        Extract desired layer from .h5 file and write to geotiff .tif

        Parameters
        ----------
        layer : str
            Layer to extract
        geotiff : str
            Path to geotiff file
        """
        self._extract_layer(self._excl_h5, layer, geotiff=geotiff,
                            hsds=self._hsds)

    @classmethod
    def layers_to_h5(cls, excl_h5, layers, chunks=(128, 128),
                     replace=True, check_tiff=True,
                     transform_atol=0.01, coord_atol=0.001,
                     descriptions=None, scale_factors=None):
        """
        Create exclusions .h5 file, or load layers into existing exclusion .h5
        file from provided geotiffs

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing or to contain exclusion layers
        layers : list | dict
            List of geotiffs to load
            or dictionary mapping goetiffs to the layers to load
        chunks : tuple, optional
            Chunk size of exclusions in Geotiff, by default (128, 128)
        replace : bool, optional
            Flag to replace existing layers if needed, by default True
        check_tiff : bool, optional
            Flag to check tiff profile and coordinates against exclusion .h5
            profile and coordinates, by default True
        transform_atol : float, optional
            Absolute tolerance parameter when comparing geotiff transform data,
            by default 0.01
        coord_atol : float, optional
            Absolute tolerance parameter when comparing new un-projected
            geotiff coordinates against previous coordinates, by default 0.001
        description : dict, optional
            Description of exclusion layers, by default None
        scale_factor : dict, optional
            Scale factors and dtypes to use when scaling given layers,
            by default None
        """
        if isinstance(layers, list):
            layers = {os.path.basename(lyr).split('.')[0]: lyr
                      for lyr in layers}

        if descriptions is None:
            descriptions = {}

        if scale_factors is None:
            scale_factors = {}

        excls = cls(excl_h5, chunks=chunks, replace=replace)
        logger.info('Creating {}'.format(excl_h5))
        for layer, geotiff in layers.items():
            logger.info('- Transfering {}'.format(layer))
            description = descriptions.get(layer, None)
            scale = scale_factors.get(layer, None)
            if scale is not None:
                scale_factor = scale['scale_factor']
                dtype = scale['dtype']
            else:
                scale_factor = None
                dtype = None

            excls.geotiff_to_layer(layer, geotiff, check_tiff=check_tiff,
                                   transform_atol=transform_atol,
                                   coord_atol=coord_atol,
                                   description=description,
                                   scale_factor=scale_factor,
                                   dtype=dtype)

    @classmethod
    def extract_layers(cls, excl_h5, layers, chunks=(128, 128),
                       hsds=False):
        """
        Extract given layers from exclusions .h5 file and save to disk
        as GeoTiffs

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing or to contain exclusion layers
        layers : dict
            Dictionary mapping layers to geotiffs to create
        chunks : tuple
            Chunk size of exclusions in .h5 and Geotiffs
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        """
        excls = cls(excl_h5, chunks=chunks, hsds=hsds)
        logger.info('Extracting layers from {}'.format(excl_h5))
        for layer, geotiff in layers.items():
            logger.info('- Extracting {}'.format(geotiff))
            excls.layer_to_geotiff(layer, geotiff)

    @classmethod
    def extract_all_layers(cls, excl_h5, out_dir, chunks=(128, 128),
                           hsds=False):
        """
        Extract all layers from exclusions .h5 file and save to disk
        as GeoTiffs

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing or to contain exclusion layers
        out_dir : str
            Path to output directory into which layers should be saved as
            GeoTiffs
        chunks : tuple
            Chunk size of exclusions in .h5 and Geotiffs
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        """
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        excls = cls(excl_h5, chunks=chunks, hsds=hsds)
        logger.info('Extracting layers from {}'.format(excl_h5))
        for layer in excls.layers:
            geotiff = os.path.join(out_dir, "{}.tif".format(layer))
            logger.info('- Extracting {}'.format(geotiff))
            excls.layer_to_geotiff(layer, geotiff)



# -*- coding: utf-8 -*-
"""
Offshore input building and extraction tools
"""
from .assembly_areas import AssemblyAreas
from .dist_to_ports import DistanceToPorts
from .offshore_inputs import OffshoreInputs



# -*- coding: utf-8 -*-
"""
setup.py
"""
import os
from codecs import open
from setuptools import setup, find_packages
from setuptools.command.develop import develop
from subprocess import check_call
import shlex
from warnings import warn

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    readme = f.read()

with open(os.path.join(here, "reVX", "version.py"), encoding="utf-8") as f:
    version = f.read()

version = version.split('=')[-1].strip().strip('"').strip("'")


class PostDevelopCommand(develop):
    """
    Class to run post setup commands
    """

    def run(self):
        """
        Run method that tries to install pre-commit hooks
        """
        try:
            check_call(shlex.split("pre-commit install"))
        except Exception as e:
            warn("Unable to run 'pre-commit install': {}"
                 .format(e))

        develop.run(self)


with open("requirements.txt") as f:
    install_requires = f.readlines()

test_requires = ["pytest>=5.2", ]
description = ("National Renewable Energy Laboratory's (NREL's) Renewable "
               "Energy Potential(V) eXchange Tool: reVX")

setup(
    name="NREL-reVX",
    version=version,
    description=description,
    long_description=readme,
    author="Michael Rossol",
    author_email="michael.rossol@nrel.gov",
    url="https://nrel.github.io/reVX/",
    packages=find_packages(),
    package_dir={"reVX": "reVX"},
    entry_points={
        "console_scripts": ["reVX=reVX.cli:main",
                            "reV-ReEDS=reVX.reeds.reeds_cli:main",
                            "reV-rpm=reVX.rpm.rpm_cli:main",
                            "reV-plexos=reVX.plexos.rev_reeds_plexos_cli:main",
                            "plexos-plants=reVX.plexos.plexos_plants_cli:main",
                            "setbacks=reVX.setbacks.setbacks_cli:main",
                            ("prominent-wind-dirs=reVX.wind_dirs."
                             "prominent_wind_dirs_cli:main"),
                            ("mean-wind-dirs=reVX.wind_dirs."
                             "mean_wind_dirs_cli:main"),
                            ("offshore-dist-to-ports=reVX.offshore."
                             "dist_to_ports_cli:main"),
                            ("offshore-assembly-areas=reVX.offshore."
                             "assembly_areas_cli:main"),
                            ("offshore-inputs=reVX.offshore."
                             "offshore_inputs_cli:main"),
                            ("turbine-flicker=reVX.turbine_flicker."
                             "turbine_flicker_cli:main"),
                            ("simple-plant-builder=reVX.plexos."
                             "simple_plant_builder_cli:main"),
                            ("xmission-cost-creator=reVX.least_cost_xmission."
                             "cost_creator_cli:main"),
                            ("least-cost-xmission=reVX.least_cost_xmission."
                             "least_cost_xmission_cli:main"),
                            ("least-cost-paths=reVX.least_cost_xmission."
                             "least_cost_paths_cli:main")
                            ],
    },
    include_package_data=True,
    license="BSD 3-Clause",
    zip_safe=False,
    keywords="reVX",
    python_requires='>=3.7,<3.9',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    test_suite="tests",
    install_requires=install_requires,
    extras_require={
        "test": test_requires,
        "dev": test_requires + ["flake8", "pre-commit", "pylint"],
    },
    cmdclass={"develop": PostDevelopCommand},
)



# -*- coding: utf-8 -*-
"""
Extract offshore inputs from exclusion layers
"""
import logging
import numpy as np
import pandas as pd
from scipy.ndimage import center_of_mass
from scipy.spatial import cKDTree
from warnings import warn

from reV.handlers.exclusions import ExclusionLayers
from reV.utilities.exceptions import MultiFileExclusionError
from reVX.utilities.utilities import log_versions, coordinate_distance
from rex.resource import Resource
from rex.utilities.utilities import parse_table, get_lat_lon_cols

logger = logging.getLogger(__name__)


class OffshoreInputs(ExclusionLayers):
    """
    Class to extract offshore inputs from offshore inputs .h5 at desired
    offshore site gids. Mapping is based on the techmapping dataset (tm_dset).
    Offshore input values are taken from the array pixel closest to the
    center of mass of each offshore site gid.
    """
    DEFAULT_INPUT_LAYERS = {
        'array_efficiency': 'aeff',
        'bathymetry': 'depth',
        'dist_to_coast': 'dist_s_to_l',
        'assembly_areas': 'dist_a_to_s',
        'ports_operations': 'dist_op_to_s',
        'ports_construction': 'dist_p_to_s',
        'ports_construction_nolimits': 'dist_p_to_s_nolimit',
        'weather_downtime_fixed_bottom': 'fixed_downtime',
        'weather_downtime_floating': 'floating_downtime',
        'weather_downtime_mean_wave_height_bouy': 'hs_average'
    }

    def __init__(self, inputs_fpath, offshore_sites, tm_dset='techmap_wtk'):
        """
        Parameters
        ----------
        inputs_fpath : str
            Path to offshore inputs .h5 file
        offshore_sites : str | list | tuple | ndarray |pandas.DataFrame
            - Path to .csv|.json file with offshore sites meta data
            - Path to a WIND Toolkit .h5 file to extact site meta from
            - List, tuple, or vector of offshore gids
            - Pre-extracted site meta DataFrame
        tm_dset : str, optional
            Dataset / layer name for wind toolkit techmap,
            by default 'techmap_wtk'
        """
        log_versions(logger)
        super().__init__(inputs_fpath)
        self._offshore_meta = self._create_offshore_meta(offshore_sites,
                                                         tm_dset)

    def __repr__(self):
        msg = "{} from {}".format(self.__class__.__name__, self.inputs_fpath)

        return msg

    @property
    def inputs_fpath(self):
        """
        .h5 file containing offshore input layers

        Returns
        -------
        str
        """
        return self.h5_file

    @property
    def meta(self):
        """
        Offshore site meta data including mapping to input layer row and column
        index

        Returns
        -------
        pandas.DataFrame
        """
        return self._offshore_meta

    @property
    def lat_lons(self):
        """
        Offshore sites coordinates (lat, lons)

        Returns
        -------
        ndarray
        """
        lat_lon_cols = get_lat_lon_cols(self.meta)

        return self.meta[lat_lon_cols].values

    @property
    def row_ids(self):
        """
        Input layer array row ids that correspond to desired offshore sites

        Returns
        -------
        ndarray
        """
        return self.meta['row_idx'].values

    @property
    def column_ids(self):
        """
        Input layer array column ids that correspond to desired offshore sites

        Returns
        -------
        ndarray
        """
        return self.meta['col_idx'].values

    @staticmethod
    def _parse_offshore_sites(offshore_sites):
        """
        Load offshore sites from disc if needed

        Parameters
        ----------
        offshore_sites : str | list | tuple | ndarray |pandas.DataFrame
            - Path to .csv|.json file with offshore sites meta data
            - Path to a WIND Toolkit .h5 file to extact site meta from
            - List, tuple, or vector of offshore gids
            - Pre-extracted site meta DataFrame

        Returns
        -------
        offshore_sites : pandas.DataFrame
            Offshore sites meta data
        """
        if isinstance(offshore_sites, str):
            if offshore_sites.endswith('.h5'):
                with Resource(offshore_sites) as f:
                    offshore_sites = f.meta
                    if offshore_sites.index.name == 'gid':
                        offshore_sites = offshore_sites.reset_index()

            else:
                offshore_sites = parse_table(offshore_sites)
        elif isinstance(offshore_sites, (tuple, list, np.ndarray)):
            offshore_sites = pd.DataFrame({'gid': offshore_sites})

        if not isinstance(offshore_sites, pd.DataFrame):
            msg = ("offshore sites must be a .csv, .json, or .h5 file path, "
                   "or a pre-extracted pandas DataFrame, but {} was provided"
                   .format(offshore_sites))
            logger.error(msg)
            raise ValueError(msg)

        if 'offshore' in offshore_sites:
            mask = offshore_sites['offshore'] == 1
            offshore_sites = offshore_sites.loc[mask]

        return offshore_sites

    @classmethod
    def _parse_input_layers(cls, input_layers=None):
        """
        Parse offshore inputs to extract from .h5 exclusion layers.
        "input_layers" can be:
        - A single layer to extract
        - A list of layers to extract
        - A dictionary with the output column name mapped to the layer to
          extract

        Parameters
        ----------
        input_layers : str | list | dict
            Input layer, list of input layers, to extract, or dictionary
            mapping the input layers to extract to the column names to save
            them under
        Returns
        -------
        dict
            Dictionary mapping the column name to the layer to extract
        """
        msg = ''
        if input_layers is None:
            input_layers = cls.DEFAULT_INPUT_LAYERS
            msg += '"input_layers" not provided, using defaults. '
        else:
            if isinstance(input_layers, str):
                input_layers = [input_layers]

            if isinstance(input_layers, (tuple, list, np.ndarray)):
                input_layers = {layer: layer for layer in input_layers}

        if not isinstance(input_layers, dict):
            msg = ('Expecting "layers" to be a the name of a single input '
                   'layer, a list of input layers, or a dictionary mapping '
                   'desired input layers to desired output column names, but '
                   'recieved: {}'.format(type(input_layers)))
            logger.error(msg)
            raise TypeError(msg)

        msg += 'Extracting {}'.format(input_layers)
        logger.info(msg)

        return input_layers

    def _preflight_multi_file(self):
        """Run simple multi-file exclusion checks."""
        lat_shape = self.h5.shapes['latitude']
        lon_shape = self.h5.shapes['longitude']
        for layer in self.layers:
            if layer not in ['assembly_areas', 'array_efficiency']:
                lshape = self.h5.shapes[layer]
                lshape = lshape[1:] if len(lshape) > 2 else lshape
                if lshape != lon_shape or lshape != lat_shape:
                    msg = ('Shape of layer "{}" is {} which does not match '
                           'latitude and longitude shapes of {} and {}. '
                           'Check your exclusion file inputs: {}'
                           .format(layer, self.h5.shapes[layer],
                                   lat_shape, lon_shape, self.h5._h5_files))
                    logger.error(msg)
                    raise MultiFileExclusionError(msg)

        check_attrs = ('height', 'width', 'crs', 'transform')
        base_profile = {}
        for fp in self.h5_file:
            with ExclusionLayers(fp) as f:
                if not base_profile:
                    base_profile = f.profile
                else:
                    for attr in check_attrs:
                        if attr not in base_profile or attr not in f.profile:
                            msg = ('Multi-file exclusion inputs from {} '
                                   'dont have profiles with height, width, '
                                   'crs, and transform: {} and {}'
                                   .format(self.h5_file, base_profile,
                                           f.profile))
                            logger.error(msg)
                            raise MultiFileExclusionError(msg)

                        if base_profile[attr] != f.profile[attr]:
                            msg = ('Multi-file exclusion inputs from {} '
                                   'dont have matching "{}": {} and {}'
                                   .format(self.h5_file, attr,
                                           base_profile[attr],
                                           f.profile[attr]))
                            logger.error(msg)
                            raise MultiFileExclusionError(msg)

    def _reduce_tech_map(self, tm_dset='techmap_wtk', offshore_gids=None,
                         offshore_dset='dist_to_coast'):
        """
        Find the row and column indices that correspond to the centriod of
        each offshore gid in exclusions layers. If offshore gids are not
        provided the centroid of every gid is in techmap.

        Parameters
        ----------
        inputs_fpath : str
            Path to offshore inputs .h5 file
        tm_dset : str, optional
            Dataset / layer name for wind toolkit techmap,
            by default 'techmap_wtk'
        offshore_gids : ndarray | list, optional
            Vector or list of offshore gids, by default None
        offshore_dset : str, optional
            Exclusions layer to differentiate between onshore and offshore
            pixels, by default 'dist_to_coast'

        Returns
        -------
        tech_map : pandas.DataFrame
            DataFrame mapping resource gid to exclusions latitude, longitude,
            row index, column index
        """

        tech_map = self[tm_dset]
        # exclude onshore pixels
        tech_map[self[offshore_dset] <= 0] = -1

        gids = np.unique(tech_map)

        if offshore_gids is None:
            offshore_gids = gids[gids != -1]
        else:
            missing = ~np.isin(offshore_gids, gids)
            if np.any(missing):
                msg = ('The following offshore gids were requested but are '
                       'not availabe in {} and will not be extracted:\n{}'
                       .format(tm_dset, offshore_gids[missing]))
                logger.warning(msg)
                warn(msg)
                offshore_gids = offshore_gids[~missing]

        # Increment techmap and gids by 1 as center of mass cannot use an
        # index of 0
        tech_map += 1
        offshore_gids += 1

        tech_map = np.array(center_of_mass(tech_map, labels=tech_map,
                                           index=offshore_gids),
                            dtype=np.uint32)

        tech_map = pd.DataFrame(tech_map, columns=['row_idx', 'col_idx'])
        tech_map['gid'] = offshore_gids - 1

        return tech_map

    def _create_offshore_meta(self, offshore_sites, tm_dset='techmap_wtk'):
        """
        Create offshore meta from offshore sites and techmap

        Parameters
        ----------
        offshore_sites : str | pandas.DataFrame
            Path to .csv file with offshore sites or offshore meta, or path
            to a .h5 file to extact site meta from, or pre-extracted site meta
            DataFrame
        tm_dset : str, optional
            Dataset / layer name for wind toolkit techmap,
            by default 'techmap_wtk'

        Returns
        -------
        offshore_meta : pandas.DataFrame
            Offshore sites meta data including mapping to input layers
        """
        offshore_sites = self._parse_offshore_sites(offshore_sites)
        if 'gid' not in offshore_sites:
            msg = ('Cannot find offshore WIND Toolkit "gid"s of interest! '
                   'Offshore sites input must have a "gid" column: {}'
                   .format(list(offshore_sites.columns)))
            logger.error(msg)
            raise RuntimeError(msg)

        offshore_gids = offshore_sites['gid'].values.astype(np.int32)
        tech_map = self._reduce_tech_map(tm_dset=tm_dset,
                                         offshore_gids=offshore_gids)

        offshore_meta = pd.merge(offshore_sites, tech_map, on='gid',
                                 how='inner')

        return offshore_meta

    def compute_assembly_dist(self, layer):
        """
        Extract the distance from ports to assembly area and then compute
        the distance from nearest assembly area to sites

        Parameters
        ----------
        layer : str
            Name of assembly area table/dataset

        Returns
        -------
        out : dict
            Dictionary containing the distance from ports to assembly areas
            ('dist_p_to_a') and distance from nearest assembly area to sites
            ('dist_a_to_s')
        """
        assembly_areas = pd.DataFrame(self.h5[layer])
        assembly_areas = self.h5.df_str_decode(assembly_areas)
        lat_lon_cols = get_lat_lon_cols(assembly_areas)
        area_coords = assembly_areas[lat_lon_cols].values.astype(np.float32)
        # pylint: disable = not-callable
        tree = cKDTree(area_coords)

        site_lat_lons = self.lat_lons
        _, pos = tree.query(site_lat_lons)
        assert len(pos) == len(site_lat_lons)

        out = {'dist_p_to_a': np.zeros(len(site_lat_lons), dtype=np.float32),
               'dist_a_to_s': np.zeros(len(site_lat_lons), dtype=np.float32)}

        for i, area in assembly_areas.iterrows():
            logger.debug("Computing distance from assembly area {} to "
                         "all offshore sites".format(area))
            a_pos = np.where(pos == i)[0]
            # extract distance from ports to assembly areas
            out['dist_p_to_a'][a_pos] = area['dist_p_to_a']

            # compute distance from assembly areas to sites
            area_coords = area[lat_lon_cols].values.astype(np.float32)
            area_coords = np.expand_dims(area_coords, 0)
            out['dist_a_to_s'][a_pos] = coordinate_distance(
                area_coords, site_lat_lons[a_pos])

        return out

    def extract_array_efficiency(self, layer):
        """
        Extract array efficiency

        Parameters
        ----------
        layer : str
            Name of array efficiency table/dataset

        Returns
        -------
        aeff : ndarray
            Vector of array efficiency 'aeff' values for desired offshore sites
        """
        aeff = pd.DataFrame(self.h5[layer])
        aeff = self.h5.df_str_decode(aeff)
        lat_lon_cols = get_lat_lon_cols(aeff)
        aeff_coords = aeff[lat_lon_cols].values.astype(np.float32)
        # pylint: disable = not-callable
        tree = cKDTree(aeff_coords)

        site_lat_lons = self.lat_lons
        _, pos = tree.query(site_lat_lons)
        assert len(pos) == len(site_lat_lons)

        aeff = aeff['aeff'].values.astype(np.float32)

        return aeff[pos]

    def extract_input_layer(self, layer):
        """
        Extract input data for desired layer

        Parameters
        ----------
        layer : str
            Desired input layer

        Returns
        -------
        data : ndarray
            Input layer data for desired offshore sites
        """
        data = self[layer, self.row_ids, self.column_ids]

        return data

    def get_offshore_inputs(self, input_layers=None,
                            downtime_tolerance=0.01):
        """
        Extract data for the desired layers

        Parameters
        ----------
        input_layers : str | list | dict
            Input layer, list of input layers, to extract, or dictionary
            mapping the input layers to extract to the column names to save
            them under, by default None
        downtime_tolerance : float
            Amount to shift weather downtime layer by so they are > 0 and/or
            < 1, by default 0.01

        Returns
        -------
        out : pandas.DataFrame
            Updated meta data table with desired layers
        """
        input_layers = self._parse_input_layers(input_layers=input_layers)

        out = self.meta.copy()
        for layer, col in input_layers.items():
            if layer not in self:
                msg = ("{} is not a valid offshore input layers, please "
                       "choice one of: {}".format(layer, self.layers))
                logger.error(msg)
                raise KeyError(msg)

            if layer.startswith('assembly'):
                for col, data in self.compute_assembly_dist(layer).items():
                    out[col] = data
            elif layer.startswith('array'):
                out[col] = self.extract_array_efficiency(layer)
            else:
                layer_out = self.extract_input_layer(layer)
                if layer == 'bathymetry':
                    layer_out = np.abs(layer_out)
                elif layer == 'dist_to_coast':
                    layer_out /= 1000
                elif layer.startswith('weather_downtime'):
                    layer_out[layer_out <= 0] = downtime_tolerance
                    if 'wave_height' not in layer:
                        layer_out[layer_out >= 1] = 1 - downtime_tolerance

                out[col] = layer_out

        return out

    @classmethod
    def extract(cls, inputs_fpath, offshore_sites, tm_dset='techmap_wtk',
                input_layers=None, downtime_tolerance=0.01, out_fpath=None):
        """
        Extract data from desired input layers for desired offshore sites

        Parameters
        ----------
        inputs_fpath : str
            Path to offshore inputs .h5 file
        offshore_sites : str | list | tuple | ndarray |pandas.DataFrame
            - Path to .csv|.json file with offshore sites meta data
            - Path to a WIND Toolkit .h5 file to extact site meta from
            - List, tuple, or vector of offshore gids
            - Pre-extracted site meta DataFrame
        tm_dset : str, optional
            Dataset / layer name for wind toolkit techmap,
            by default 'techmap_wtk'
        input_layers : str | list | dict
            Input layer, list of input layers, to extract, or dictionary
            mapping the input layers to extract to the column names to save
            them under, by default None
        downtime_tolerance : float
            Amount to shift weather downtime layer by so they are > 0 and/or
            < 1, by default 0.01
        out_fpath : str, optional
            Output .csv path to save offshore inputs too, by default None

        Returns
        -------
        out : pandas.DataFrame
            Updated meta data table with desired layers
        """
        with cls(inputs_fpath, offshore_sites, tm_dset=tm_dset) as off_ipt:
            out = off_ipt.get_offshore_inputs(
                input_layers=input_layers,
                downtime_tolerance=downtime_tolerance)

        if out_fpath:
            out.to_csv(out_fpath, index=False)

        return out



# -*- coding: utf-8 -*-
"""
Compute setbacks exclusions
"""
import logging
import os
from warnings import warn
import geopandas as gpd

from rex.utilities import log_mem

from reVX.setbacks.base import BaseSetbacks


logger = logging.getLogger(__name__)


class ParcelSetbacks(BaseSetbacks):
    """
    Parcel setbacks, using negative buffers.
    """

    def get_regulation_setback(self, county_regulations):
        """Compute the setback distance for the county.

        Compute the setback distance (in meters) from the
        county regulations or the plant height.

        Parameters
        ----------
        county_regulations : pandas.Series
            Pandas Series with regulations for a single county
            or feature type. At a minimum, this Series must
            contain the following columns: `Value Type`, which
            specifies wether the value is a multiplier or static height,
            `Value`, which specifies the numeric value of the setback or
            multiplier. Valid options for the `Value Type` are:
                - "Structure Height multiplier"
                - "Meters"

        Returns
        -------
        setback : float | None
            Setback distance in meters, or `None` if the setback
            `Value Type` was not recognized.
        """

        setback_type = county_regulations["Value Type"].strip()
        setback = float(county_regulations["Value"])
        if setback_type.lower() == "structure height multiplier":
            setback *= self.base_setback_dist
        elif setback_type.lower() != "meters":
            msg = ("Cannot create setback for {}, expecting "
                   '"Maximum Structure Height", or '
                   '"Meters", but got {}'
                   .format(county_regulations["County"], setback_type))
            logger.warning(msg)
            warn(msg)
            setback = None

        return setback

    def compute_generic_setbacks(self, features_fpath):
        """Compute generic setbacks.

        This method will compute the setbacks using a generic setback
        of `base_setback_dist * multiplier`.

        Parameters
        ----------
        features_fpath : str
            Path to shape file with features to compute setbacks from.

        Returns
        -------
        setbacks : ndarray
            Raster array of setbacks
        """
        logger.info("Computing generic setbacks")
        setback_features = self._parse_features(features_fpath)

        setbacks = [
            (geom, 1) for geom in setback_features.buffer(0).difference(
                setback_features.buffer(-1 * self.generic_setback)
            )
        ]

        return self._rasterize_setbacks(setbacks)

    @staticmethod
    def _compute_local_setbacks(features, cnty, setback):
        """Compute local features setbacks.

        This method will compute the setbacks using a county-specific
        regulations file that specifies either a static setback or a
        multiplier value that will be used along with the plant height
        to compute the setback.

        Parameters
        ----------
        features : geopandas.GeoDataFrame
            Features to setback from.
        cnty : geopandas.GeoDataFrame
            Regulations for a single county.
        setback : int
            Setback distance in meters.

        Returns
        -------
        setbacks : list
            List of setback geometries.
        """
        logger.debug('- Computing setbacks for county FIPS {}'
                     .format(cnty.iloc[0]['FIPS']))
        log_mem(logger)
        mask = features.centroid.within(cnty['geometry'].values[0])
        tmp = features.loc[mask]

        setbacks = [
            (geom, 1) for geom in tmp.buffer(0).difference(
                tmp.buffer(-1 * setback)
            )
        ]

        return setbacks

    def _parse_regulations(self, regulations_fpath):
        """
        Parse parcel regulations, reduce table to just property lines

        Parameters
        ----------
        regulations_fpath : str
            Path to parcel regulations .csv file

        Returns
        -------
        regulations : pandas.DataFrame
            Parcel regulations table
        """
        regulations = super()._parse_regulations(regulations_fpath)

        mask = regulations['Feature Type'].apply(str.strip) == 'Property Line'
        regulations = regulations.loc[mask]

        return regulations

    def _check_regulations(self, features_fpath):
        """
        Reduce regs to state corresponding to features_fpath if needed.

        Parameters
        ----------
        features_fpath : str
            Path to shape file with features to compute setbacks from.
            This file needs to have the state in the filename.

        Returns
        -------
        regulations : geopandas.GeoDataFrame
            Parcel regulations
        """
        state = os.path.basename(features_fpath).split('.')[0]
        state = ''.join(filter(str.isalpha, state.lower()))

        regulation_states = self.regulations.State.apply(
            lambda s: ''.join(filter(str.isalpha, s.lower()))
        )

        mask = regulation_states == state
        regulations = self.regulations[mask].reset_index(drop=True)

        logger.debug(
            'Computing setbacks for parcel regulations in {} counties'
            .format(len(regulations))
        )

        return regulations

    def _parse_features(self, features_fpath):
        """Abstract method to parse features.

        Parameters
        ----------
        features_fpath : str
            Path to file containing features to setback from.

        Returns
        -------
        `geopandas.GeoDataFrame`
            Geometries of features to setback from in exclusion
            coordinate system.
        """
        features = gpd.read_file(features_fpath)
        if features.crs is None:
            features = features.set_crs("EPSG:4326")
        return features.to_crs(crs=self.crs)

    @classmethod
    def run(cls, excl_fpath, parcels_path, out_dir, base_setback_dist,
            regulations_fpath=None, multiplier=None,
            chunks=(128, 128), max_workers=None, replace=False, hsds=False):
        """
        Compute parcel setbacks and write them to a geotiff.
        If a regulations file is given, compute local setbacks,
        otherwise compute generic setbacks using the given multiplier
        and the plant height.

        Parameters
        ----------
        excl_fpath : str
            Path to .h5 file containing exclusion layers, will also be
            the location of any new setback layers.
        parcels_path : str
            Path to parcels file or directory containing parcel files.
            This path can contain any pattern that can be used in the
            glob function. For example, `/path/to/features/[A]*` would
            match with all the features in the direcotry
            `/path/to/features/` that start with "A". This input
            can also be a directory, but that directory must ONLY
            contain feature files. If your feature files are mixed
            with other files or directories, use something like
            `/path/to/features/*.geojson`.
        out_dir : str
            Directory to save setbacks geotiff(s) into
        base_setback_dist : float | int
            Base setback distance (m). This value will be used to
            calculate the setback distance when a multiplier is provided
            either via the `regulations_fpath`csv or the `multiplier`
            input. In this case, the setbacks will be calculated using
            `base_setback_dist * multiplier`.
        regulations_fpath : str | None, optional
            Path to regulations .csv file. At a minimum, this csv must
            contain the following columns: `Value Type`, which
            specifies wether the value is a multiplier or static height,
            `Value`, which specifies the numeric value of the setback or
            multiplier, and `FIPS`, which specifies a unique 5-digit
            code for each county (this can be an integer - no leading
            zeros required). Typically, this csv will also have a
            `Feature Type` column that labels the type of setback
            that each row represents. Valid options for the `Value Type`
            are:
                - "Structure Height multiplier"
                - "Meters"
            If this input is `None`, a generic setback of
            `base_setback_dist * multiplier` is used. By default `None`.
        multiplier : int | float | str | None, optional
            Setback multiplier to use if regulations are not supplied.
            It is multiplied with plant height to calculate the setback
            distance. If supplied along with `regulations_fpath`, this
            input will be ignored. By default `None`.
        chunks : tuple, optional
            Chunk size to use for setback layers, if None use default
            chunk size in excl_fpath, By default `(128, 128)`.
        max_workers : int, optional
            Number of workers to use for setback computation, if 1 run
            in serial, if > 1 run in parallel with that many workers,
            if `None`, run in parallel on all available cores.
            By default `None`.
        replace : bool, optional
            Flag to replace geotiff if it already exists.
            By default `False`.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on
            AWS behind HSDS. By default `False`.
        """
        setbacks = cls(excl_fpath, base_setback_dist=base_setback_dist,
                       regulations_fpath=regulations_fpath,
                       multiplier=multiplier,
                       hsds=hsds, chunks=chunks)

        parcels_path = setbacks._get_feature_paths(parcels_path)
        for fpath in parcels_path:
            geotiff = os.path.basename(fpath).split('.')[0]
            geotiff += '.tif'
            geotiff = os.path.join(out_dir, geotiff)
            if os.path.exists(geotiff) and not replace:
                msg = ('{} already exists, setbacks will not be re-computed '
                       'unless replace=True'.format(geotiff))
                logger.error(msg)
            else:
                logger.info("Computing setbacks from parcels in {} and saving "
                            "to {}".format(fpath, geotiff))
                setbacks.compute_setbacks(fpath, geotiff=geotiff,
                                          max_workers=max_workers,
                                          replace=replace)



# -*- coding: utf-8 -*-
"""
Distance to Ports Command Line Interface
"""
import click
import logging
import os

from rex.utilities.loggers import init_mult
from rex.utilities.cli_dtypes import STR, INT
from rex.utilities.hpc import SLURM
from rex.utilities.utilities import get_class_properties

from reVX.config.offshore import DistToPortsConfig
from reVX.offshore.dist_to_ports import DistanceToPorts
from reVX import __version__

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('--name', '-n', default='DistToPorts', type=STR,
              show_default=True,
              help='Job name.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, verbose):
    """
    Distance to Ports Command Line Interface
    """
    ctx.ensure_object(dict)
    ctx.obj['VERBOSE'] = verbose
    ctx.obj['NAME'] = name


@main.command()
def valid_config_keys():
    """
    Echo the valid Distance to Port config keys
    """
    click.echo(', '.join(get_class_properties(DistToPortsConfig)))


def run_local(ctx, config):
    """
    Compute distance to ports locally using config

    Parameters
    ----------
    ctx : click.ctx
        click ctx object
    config : reVX.config.dist_to_ports.DistToPortsConfig
        Distance to ports config object.
    """
    ctx.obj['NAME'] = config.name
    ctx.invoke(local,
               ports_fpath=config.ports_fpath,
               excl_fpath=config.excl_fpath,
               out_dir=config.dirout,
               input_dist_layer=config.input_dist_layer,
               max_workers=config.execution_control.max_workers,
               replace=config.replace,
               log_dir=config.log_directory,
               verbose=config.log_level)


@main.command()
@click.option('--config', '-c', required=True,
              type=click.Path(exists=True),
              help='Filepath to DistToPorts config json file.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config, verbose):
    """
    Compute distance to ports from a config.
    """

    config = DistToPortsConfig(config)

    if 'VERBOSE' in ctx.obj:
        if any((ctx.obj['VERBOSE'], verbose)):
            config._log_level = logging.DEBUG
    elif verbose:
        config._log_level = logging.DEBUG

    if config.execution_control.option == 'local':
        run_local(ctx, config)

    if config.execution_control.option == 'eagle':
        eagle(config)


@main.command()
@click.option('--ports_fpath', '-ports', required=True,
              type=click.Path(exists=True),
              help=("Path to shape, csv, or json file containing ports to "
                    "compute least cost distance to"))
@click.option('--excl_fpath', '-excl', required=True,
              type=click.Path(exists=True),
              help="Filepath to exclusions h5 with techmap dataset.")
@click.option('--out_dir', '-o', required=True, type=click.Path(),
              help='Directory to save distance to port geotiffs too.')
@click.option('--input_dist_layer', '-idl', default='dist_to_coast',
              show_default=True,
              help=("Exclusions layer with distance to coast values"))
@click.option('--max_workers', '-mw', default=1, type=INT,
              show_default=True,
              help=(" Number of workers to use for setback computation, if 1 "
                    "run in serial, if > 1 run in parallel with that many "
                    "workers, if None run in parallel on all available cores"))
@click.option('--replace', '-rm', is_flag=True,
              help="Flag to replace existing ports geotiffs")
@click.option('--log_dir', '-log', default=None, type=STR,
              show_default=True,
              help='Directory to dump log files.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def local(ctx, ports_fpath, excl_fpath, out_dir, input_dist_layer, max_workers,
          replace, log_dir, verbose):
    """
    Compute distance to ports on local hardware
    """
    name = ctx.obj['NAME']
    if 'VERBOSE' in ctx.obj:
        verbose = any((ctx.obj['VERBOSE'], verbose))

    log_modules = [__name__, 'reVX', 'reV', 'rex']
    init_mult(name, log_dir, modules=log_modules, verbose=verbose)

    logger.info('Computing distance to ports in {} \n'
                'Outputs to be stored in: {}'.format(ports_fpath, excl_fpath))

    DistanceToPorts.run(ports_fpath, excl_fpath, out_dir,
                        input_dist_layer=input_dist_layer,
                        max_workers=max_workers,
                        replace=replace)


def get_node_cmd(config):
    """
    Get the node CLI call for distance to ports computation.

    Parameters
    ----------
    config : reVX.config.dist_to_ports.DistToPortsConfig
        Distance to ports config object.

    Returns
    -------
    cmd : str
        CLI call to submit to SLURM execution.
    """

    args = ['-n {}'.format(SLURM.s(config.name)),
            'local',
            '-ports {}'.format(SLURM.s(config.ports_fpath)),
            '-excl {}'.format(SLURM.s(config.excl_fpath)),
            '-o {}'.format(SLURM.s(config.dirout)),
            '-idl {}'.format(SLURM.s(config.input_dist_layer)),
            '-odl {}'.format(SLURM.s(config.output_dist_layer)),
            '-mw {}'.format(SLURM.s(config.execution_control.max_workers)),
            '-log {}'.format(SLURM.s(config.log_directory)),
            ]

    if config.replace:
        args.append('-r')

    if config.log_level == logging.DEBUG:
        args.append('-v')

    cmd = ('python -m reVX.offshore.dist_to_ports_cli {}'
           .format(' '.join(args)))
    logger.debug('Submitting the following cli call:\n\t{}'.format(cmd))

    return cmd


def eagle(config):
    """
    Compute distance to ports on Eagle HPC.

    Parameters
    ----------
    config : reVX.config.dist_to_ports.DistToPortsConfig
        Distance to ports config object.
    """

    cmd = get_node_cmd(config)
    name = config.name
    log_dir = config.log_directory
    stdout_path = os.path.join(log_dir, 'stdout/')

    slurm_manager = SLURM()

    logger.info('Compute distance to ports on Eagle with '
                'node name "{}"'.format(name))
    out = slurm_manager.sbatch(cmd,
                               alloc=config.execution_control.allocation,
                               memory=config.execution_control.memory,
                               walltime=config.execution_control.walltime,
                               feature=config.execution_control.feature,
                               name=name, stdout_path=stdout_path,
                               conda_env=config.execution_control.conda_env,
                               module=config.execution_control.module)[0]
    if out:
        msg = ('Kicked off distance to ports calculation "{}" '
               '(SLURM jobid #{}) on Eagle.'
               .format(name, out))
    else:
        msg = ('Was unable to kick off distance to ports calculation '
               '"{}". Please see the stdout error messages'
               .format(name))

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running distance to ports CLI')
        raise
